#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: December 29, 2020
#
"""It is utilities for Gaussian process regression and
Student-:math:`t` process regression."""

import numpy as np

from bayeso.utils import utils_common
from bayeso import constants


@utils_common.validate_types
def get_prior_mu(prior_mu: constants.TYPING_UNION_CALLABLE_NONE, X: np.ndarray) -> np.ndarray:
    """
    It computes the prior mean function values over inputs X.

    :param prior_mu: prior mean function or None.
    :type prior_mu: function or NoneType
    :param X: inputs for prior mean function. Shape: (n, d) or (n, m, d).
    :type X: numpy.ndarray

    :returns: zero array, or array of prior mean function values. Shape: (n, 1).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert callable(prior_mu) or prior_mu is None
    assert len(X.shape) == 2 or len(X.shape) == 3

    if prior_mu is None:
        prior_mu_X = np.zeros((X.shape[0], 1))
    else:
        prior_mu_X = prior_mu(X)
        assert len(prior_mu_X.shape) == 2
        assert X.shape[0] == prior_mu_X.shape[0]
    return prior_mu_X

@utils_common.validate_types
def validate_common_args(X_train: np.ndarray, Y_train: np.ndarray,
    str_cov: str, prior_mu: constants.TYPING_UNION_CALLABLE_NONE,
    debug: bool,
    X_test: constants.TYPING_UNION_ARRAY_NONE=None,
) -> constants.TYPE_NONE:
    """
    It validates the common arguments for various functions.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param prior_mu: None, or prior mean function.
    :type prior_mu: NoneType, or function
    :param debug: flag for printing log messages.
    :type debug: bool.
    :param X_test: inputs or None. Shape: (l, d) or (l, m, d).
    :type X_test: numpy.ndarray, or NoneType, optional

    :returns: None.
    :rtype: NoneType

    :raises: AssertionError

    """

    assert isinstance(X_train, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(str_cov, str)
    assert callable(prior_mu) or prior_mu is None
    assert isinstance(debug, bool)
    assert len(X_train.shape) == 2 or len(X_train.shape) == 3
    assert len(Y_train.shape) == 2
    assert X_train.shape[0] == Y_train.shape[0]
    assert isinstance(X_test, (np.ndarray, type(None)))

    if X_test is not None:
        if len(X_train.shape) == 2:
            assert X_train.shape[1] == X_test.shape[1]
        else:
            assert X_train.shape[2] == X_test.shape[2]



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: September 24, 2020
#
"""It defines wrappers for Bayesian optimization."""

import time
import numpy as np

from bayeso import bo
from bayeso import constants
from bayeso.utils import utils_bo
from bayeso.utils import utils_common
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('wrappers_bo')


@utils_common.validate_types
def run_single_round_with_all_initial_information(model_bo: bo.BO,
    fun_target: constants.TYPING_CALLABLE,
    X_train: np.ndarray, Y_train: np.ndarray,
    num_iter: int,
    str_sampling_method_ao: str=constants.STR_SAMPLING_METHOD_AO,
    num_samples_ao: int=constants.NUM_SAMPLES_AO,
    str_mlm_method: str=constants.STR_MLM_METHOD
) -> constants.TYPING_TUPLE_FIVE_ARRAYS:
    """
    It optimizes `fun_target` for `num_iter` iterations with given `model_bo`.
    It returns the optimization results and execution times.

    :param model_bo: Bayesian optimization model.
    :type model_bo: bayeso.bo.BO
    :param fun_target: a target function.
    :type fun_target: callable
    :param X_train: initial inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: initial outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param num_iter: the number of iterations for Bayesian optimization.
    :type num_iter: int.
    :param str_sampling_method_ao: the name of initialization method for
        acquisition function optimization.
    :type str_sampling_method_ao: str., optional
    :param num_samples_ao: the number of samples for acquisition function
        optimization. If L-BFGS-B is used as an acquisition function
        optimization method, it is employed.
    :type num_samples_ao: int., optional
    :param str_mlm_method: the name of marginal likelihood maximization
        method for Gaussian process regression.
    :type str_mlm_method: str., optional

    :returns: tuple of acquired examples, their function values, overall
        execution times per iteration, execution time consumed in Gaussian
        process regression, and execution time consumed in acquisition
        function optimization. Shape: ((n + `num_iter`, d), (n + `num_iter`, 1),
        (`num_iter`, ), (`num_iter`, ), (`num_iter`, )), or ((n + `num_iter`, m, d),
        (n + `num_iter`, m, 1), (`num_iter`, ), (`num_iter`, ), (`num_iter`, )).
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(model_bo, bo.BO)
    assert callable(fun_target)
    assert isinstance(X_train, np.ndarray)
    assert isinstance(Y_train, np.ndarray)
    assert isinstance(num_iter, int)
    assert isinstance(str_sampling_method_ao, str)
    assert isinstance(num_samples_ao, int)
    assert isinstance(str_mlm_method, str)
    assert len(X_train.shape) == 2
    assert len(Y_train.shape) == 2
    assert X_train.shape[0] == Y_train.shape[0]
    assert Y_train.shape[1] == 1
    assert str_mlm_method in constants.ALLOWED_MLM_METHOD

    time_start = time.time()

    X_final = X_train
    Y_final = Y_train
    time_all_final = []
    time_surrogate_final = []
    time_acq_final = []
    for ind_iter in range(0, num_iter):
        logger.info('Iteration %d', ind_iter + 1)
        time_iter_start = time.time()

        next_point, dict_info = model_bo.optimize(X_final, Y_final,
            str_sampling_method=str_sampling_method_ao,
            num_samples=num_samples_ao, str_mlm_method=str_mlm_method)
        next_points = dict_info['next_points']
        acquisitions = dict_info['acquisitions']
        time_surrogate = dict_info['time_surrogate']
        time_acq = dict_info['time_acq']

        if model_bo.debug:
            logger.debug('next_point: %s', utils_logger.get_str_array(next_point))

        if np.where(np.linalg.norm(next_point - X_final, axis=1)\
            < constants.TOLERANCE_DUPLICATED_ACQ)[0].shape[0] > 0: # pragma: no cover
            next_point = utils_bo.get_next_best_acquisition(next_points, acquisitions, X_final)
            if model_bo.debug:
                logger.debug('next_point is repeated, so next best is selected.\
                    next_point: %s', utils_logger.get_str_array(next_point))
        X_final = np.vstack((X_final, next_point))

        time_to_evaluate_start = time.time()
        Y_final = np.vstack((Y_final, fun_target(next_point)))
        time_to_evaluate_end = time.time()
        if model_bo.debug:
            logger.debug('time consumed to evaluate: %.4f sec.',
                time_to_evaluate_end - time_to_evaluate_start)

        time_iter_end = time.time()
        time_all_final.append(time_iter_end - time_iter_start)
        time_surrogate_final.append(time_surrogate)
        time_acq_final.append(time_acq)

    time_end = time.time()

    if model_bo.debug:
        logger.debug('overall time consumed in single BO round: %.4f sec.', time_end - time_start)

    time_all_final = np.array(time_all_final)
    time_surrogate_final = np.array(time_surrogate_final)
    time_acq_final = np.array(time_acq_final)
    return X_final, Y_final, time_all_final, time_surrogate_final, time_acq_final

@utils_common.validate_types
def run_single_round_with_initial_inputs(model_bo: bo.BO,
    fun_target: constants.TYPING_CALLABLE,
    X_train: np.ndarray, num_iter: int,
    str_sampling_method_ao: str=constants.STR_SAMPLING_METHOD_AO,
    num_samples_ao: int=constants.NUM_SAMPLES_AO,
    str_mlm_method: str=constants.STR_MLM_METHOD,
) -> constants.TYPING_TUPLE_FIVE_ARRAYS:
    """
    It optimizes `fun_target` for `num_iter` iterations with given
    `model_bo` and initial inputs `X_train`.
    It returns the optimization results and execution times.

    :param model_bo: Bayesian optimization model.
    :type model_bo: bayeso.bo.BO
    :param fun_target: a target function.
    :type fun_target: callable
    :param X_train: initial inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param num_iter: the number of iterations for Bayesian optimization.
    :type num_iter: int.
    :param str_sampling_method_ao: the name of initialization method for
        acquisition function optimization.
    :type str_sampling_method_ao: str., optional
    :param num_samples_ao: the number of samples for acquisition function
        optimization. If L-BFGS-B is used as an acquisition function
        optimization method, it is employed.
    :type num_samples_ao: int., optional
    :param str_mlm_method: the name of marginal likelihood maximization
        method for Gaussian process regression.
    :type str_mlm_method: str., optional

    :returns: tuple of acquired examples, their function values, overall
        execution times per iteration, execution time consumed in Gaussian
        process regression, and execution time consumed in acquisition
        function optimization. Shape: ((n + `num_iter`, d), (n + `num_iter`, 1),
        (n + `num_iter`, ), (`num_iter`, ), (`num_iter`, )), or ((n + `num_iter`, m, d),
        (n + `num_iter`, m, 1), (n + `num_iter`, ), (`num_iter`, ), (`num_iter`, )).
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(model_bo, bo.BO)
    assert callable(fun_target)
    assert isinstance(X_train, np.ndarray)
    assert isinstance(num_iter, int)
    assert isinstance(str_sampling_method_ao, str)
    assert isinstance(num_samples_ao, int)
    assert isinstance(str_mlm_method, str)
    assert len(X_train.shape) == 2
    assert str_mlm_method in constants.ALLOWED_MLM_METHOD

    Y_train = []
    time_initials = []
    for elem in X_train:
        time_initial_start = time.time()
        Y_train.append(fun_target(elem))
        time_initial_end = time.time()
        time_initials.append(time_initial_end - time_initial_start)
    time_initials = np.array(time_initials)

    Y_train = np.array(Y_train)
    Y_train = np.reshape(Y_train, (Y_train.shape[0], 1))
    X_final, Y_final, time_all_final, time_surrogate_final, time_acq_final \
        = run_single_round_with_all_initial_information(
            model_bo,
            fun_target,
            X_train,
            Y_train,
            num_iter,
            str_sampling_method_ao=str_sampling_method_ao,
            num_samples_ao=num_samples_ao,
            str_mlm_method=str_mlm_method
        )
    return X_final, Y_final, \
        np.concatenate((time_initials, time_all_final)), \
        time_surrogate_final, time_acq_final

@utils_common.validate_types
def run_single_round(model_bo: bo.BO, fun_target: constants.TYPING_CALLABLE,
    num_init: int, num_iter: int,
    str_initial_method_bo: str=constants.STR_INITIALIZING_METHOD_BO,
    str_sampling_method_ao: str=constants.STR_SAMPLING_METHOD_AO,
    num_samples_ao: int=constants.NUM_SAMPLES_AO,
    str_mlm_method: str=constants.STR_MLM_METHOD,
    seed: constants.TYPING_UNION_INT_NONE=None
) -> constants.TYPING_TUPLE_FIVE_ARRAYS:
    """
    It optimizes `fun_target` for `num_iter` iterations with given
    `model_bo` and `num_init` initial examples.
    Initial examples are sampled by `get_initials` method in `model_bo`.
    It returns the optimization results and execution times.

    :param model_bo: Bayesian optimization model.
    :type model_bo: bayeso.bo.BO
    :param fun_target: a target function.
    :type fun_target: callable
    :param num_init: the number of initial examples for Bayesian optimization.
    :type num_init: int.
    :param num_iter: the number of iterations for Bayesian optimization.
    :type num_iter: int.
    :param str_initial_method_bo: the name of initialization method for
        sampling initial examples in Bayesian optimization.
    :type str_initial_method_bo: str., optional
    :param str_sampling_method_ao: the name of initialization method for
        acquisition function optimization.
    :type str_sampling_method_ao: str., optional
    :param num_samples_ao: the number of samples for acquisition function
        optimization. If L-BFGS-B is used as an acquisition function
        optimization method, it is employed.
    :type num_samples_ao: int., optional
    :param str_mlm_method: the name of marginal likelihood maximization
        method for Gaussian process regression.
    :type str_mlm_method: str., optional
    :param seed: None, or random seed.
    :type seed: NoneType or int., optional

    :returns: tuple of acquired examples, their function values, overall
        execution times per iteration, execution time consumed in Gaussian
        process regression, and execution time consumed in acquisition
        function optimization. Shape: ((`num_init` + `num_iter`, d),
        (`num_init` + `num_iter`, 1), (`num_init` + `num_iter`, ), (`num_iter`, ),
        (`num_iter`, )), or ((`num_init` + `num_iter`, m, d), (`num_init` + `num_iter`, m, 1),
        (`num_init` + `num_iter`, ), (`num_iter`, ), (`num_iter`, )),
        where d is a dimensionality of the problem we are solving and m is
        a cardinality of sets.
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(model_bo, bo.BO)
    assert callable(fun_target)
    assert isinstance(num_init, int)
    assert isinstance(num_iter, int)
    assert isinstance(str_initial_method_bo, str)
    assert isinstance(str_sampling_method_ao, str)
    assert isinstance(num_samples_ao, int)
    assert isinstance(str_mlm_method, str)
    assert isinstance(seed, (int, type(None)))
    assert str_initial_method_bo in constants.ALLOWED_INITIALIZING_METHOD_BO
    assert str_mlm_method in constants.ALLOWED_MLM_METHOD

    logger.info('range_X:\n%s', utils_logger.get_str_array(model_bo.range_X))
    logger.info('str_cov: %s', model_bo.str_cov)
    logger.info('str_acq: %s', model_bo.str_acq)
    logger.info('str_optimizer_method_gp: %s', model_bo.str_optimizer_method_gp)
    logger.info('str_optimizer_method_bo: %s', model_bo.str_optimizer_method_bo)
    logger.info('str_modelselection_method: %s', model_bo.str_modelselection_method)
    logger.info('num_init: %d', num_init)
    logger.info('num_iter: %d', num_iter)
    logger.info('str_initial_method_bo: %s', str_initial_method_bo)
    logger.info('str_sampling_method_ao: %s', str_sampling_method_ao)
    logger.info('num_samples_ao: %d', num_samples_ao)
    logger.info('str_mlm_method: %s', str_mlm_method)
    logger.info('seed: %s', seed)

    time_start = time.time()

    X_init = model_bo.get_initials(str_initial_method_bo, num_init, seed=seed)
    if model_bo.debug:
        logger.debug('X_init:\n%s', utils_logger.get_str_array(X_init))

    X_final, Y_final, time_all_final, time_surrogate_final, time_acq_final \
        = run_single_round_with_initial_inputs(
            model_bo, fun_target, X_init, num_iter,
            str_sampling_method_ao=str_sampling_method_ao,
            num_samples_ao=num_samples_ao,
            str_mlm_method=str_mlm_method
        )

    time_end = time.time()

    if model_bo.debug:
        logger.debug('overall time consumed including initializations: %.4f sec.',
            time_end - time_start)

    return X_final, Y_final, time_all_final, time_surrogate_final, time_acq_final



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: July 8, 2021
#
"""It defines covariance functions and their associated functions.
Derivatives of covariance functions with respect to hyperparameters are described
in `these notes <http://jungtaek.github.io/notes/derivatives_kernels.pdf>`_.
"""

import numpy as np
import scipy.spatial.distance as scisd
import scipy.linalg

from bayeso import constants
from bayeso.utils import utils_covariance
from bayeso.utils import utils_common


@utils_common.validate_types
def choose_fun_cov(str_cov: str) -> constants.TYPING_CALLABLE:
    """
    It chooses a covariance function.

    :param str_cov: the name of covariance function.
    :type str_cov: str.

    :returns: covariance function.
    :rtype: callable

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)

    if str_cov in ('eq', 'se'):
        fun_cov = cov_se
    elif str_cov == 'matern32':
        fun_cov = cov_matern32
    elif str_cov == 'matern52':
        fun_cov = cov_matern52
    else:
        raise NotImplementedError('choose_fun_cov: allowed str_cov condition,\
            but it is not implemented.')
    return fun_cov

@utils_common.validate_types
def choose_fun_grad_cov(str_cov: str) -> constants.TYPING_CALLABLE:
    """
    It chooses a function for computing gradients of covariance function.

    :param str_cov: the name of covariance function.
    :type str_cov: str.

    :returns: function for computing gradients of covariance function.
    :rtype: callable

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)

    if str_cov in ('eq', 'se'):
        fun_grad_cov = grad_cov_se
    elif str_cov == 'matern32':
        fun_grad_cov = grad_cov_matern32
    elif str_cov == 'matern52':
        fun_grad_cov = grad_cov_matern52
    else:
        raise NotImplementedError('choose_fun_grad_cov: allowed str_cov condition,\
            but it is not implemented.')
    return fun_grad_cov

@utils_common.validate_types
def get_kernel_inverse(X_train: np.ndarray, hyps: dict, str_cov: str,
    fix_noise: bool=constants.FIX_GP_NOISE,
    use_gradient: bool=False,
    debug: bool=False
) -> constants.TYPING_TUPLE_THREE_ARRAYS:
    """
    This function computes a kernel inverse without any matrix decomposition techniques.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param hyps: dictionary of hyperparameters for Gaussian process.
    :type hyps: dict.
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param use_gradient: flag for computing and returning gradients of
        negative log marginal likelihood.
    :type use_gradient: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: a tuple of kernel matrix over `X_train`, kernel matrix
        inverse, and gradients of kernel matrix. If `use_gradient` is False,
        gradients of kernel matrix would be None.
    :rtype: tuple of (numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(X_train, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(str_cov, str)
    assert isinstance(use_gradient, bool)
    assert isinstance(fix_noise, bool)
    assert isinstance(debug, bool)
    utils_covariance.check_str_cov('get_kernel_inverse', str_cov, X_train.shape)

    cov_X_X = cov_main(str_cov, X_train, X_train, hyps, True) \
        + hyps['noise']**2 * np.eye(X_train.shape[0])
    cov_X_X = (cov_X_X + cov_X_X.T) / 2.0
    inv_cov_X_X = np.linalg.inv(cov_X_X)

    if use_gradient:
        grad_cov_X_X = grad_cov_main(str_cov, X_train, X_train,
            hyps, fix_noise, same_X_Xp=True)
    else:
        grad_cov_X_X = None

    return cov_X_X, inv_cov_X_X, grad_cov_X_X

@utils_common.validate_types
def get_kernel_cholesky(X_train: np.ndarray, hyps: dict, str_cov: str,
    fix_noise: bool=constants.FIX_GP_NOISE,
    use_gradient: bool=False,
    debug: bool=False
) -> constants.TYPING_TUPLE_THREE_ARRAYS:
    """
    This function computes a kernel inverse with Cholesky decomposition.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param hyps: dictionary of hyperparameters for Gaussian process.
    :type hyps: dict.
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param use_gradient: flag for computing and returning gradients of
        negative log marginal likelihood.
    :type use_gradient: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: a tuple of kernel matrix over `X_train`, lower matrix computed
        by Cholesky decomposition, and gradients of kernel matrix. If
        `use_gradient` is False, gradients of kernel matrix would be None.
    :rtype: tuple of (numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(X_train, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(str_cov, str)
    assert isinstance(fix_noise, bool)
    assert isinstance(use_gradient, bool)
    assert isinstance(debug, bool)
    utils_covariance.check_str_cov('get_kernel_cholesky', str_cov, X_train.shape)

    cov_X_X = cov_main(str_cov, X_train, X_train, hyps, True) \
        + hyps['noise']**2 * np.eye(X_train.shape[0])
    cov_X_X = (cov_X_X + cov_X_X.T) / 2.0
    try:
        lower = scipy.linalg.cholesky(cov_X_X, lower=True)
    except np.linalg.LinAlgError: # pragma: no cover
        cov_X_X += 1e-2 * np.eye(X_train.shape[0])
        lower = scipy.linalg.cholesky(cov_X_X, lower=True)

    if use_gradient:
        grad_cov_X_X = grad_cov_main(str_cov, X_train, X_train,
            hyps, fix_noise, same_X_Xp=True)
    else:
        grad_cov_X_X = None
    return cov_X_X, lower, grad_cov_X_X

@utils_common.validate_types
def cov_se(X: np.ndarray, Xp: np.ndarray, lengthscales: constants.TYPING_UNION_ARRAY_FLOAT,
    signal: float
) -> np.ndarray:
    """
    It computes squared exponential kernel over `X` and `Xp`, where
    `lengthscales` and `signal` are given.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param lengthscales: length scales. Shape: (d, ) or ().
    :type lengthscales: numpy.ndarray, or float
    :param signal: coefficient for signal.
    :type signal: float

    :returns: kernel values over `X` and `Xp`. Shape: (n, m).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(lengthscales, (np.ndarray, float))
    assert isinstance(signal, float)
    assert len(X.shape) == 2
    assert len(Xp.shape) == 2
    if isinstance(lengthscales, np.ndarray):
        assert X.shape[1] == Xp.shape[1] == lengthscales.shape[0]
    else:
        assert X.shape[1] == Xp.shape[1]
    dist = scisd.cdist(X / lengthscales, Xp / lengthscales, metric='euclidean')
    cov_X_Xp = signal**2 * np.exp(-0.5 * dist**2)
    return cov_X_Xp

@utils_common.validate_types
def grad_cov_se(cov_X_Xp: np.ndarray, X: np.ndarray, Xp: np.ndarray, hyps: dict,
    num_hyps: int, fix_noise: bool
) -> np.ndarray:
    """
    It computes gradients of squared exponential kernel over `X` and `Xp`,
    where `hyps` is given.

    :param cov_X_Xp: covariance matrix. Shape: (n, m).
    :type cov_X_Xp: numpy.ndarray
    :param X: one inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param num_hyps: the number of hyperparameters == l.
    :type num_hyps: int.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool.

    :returns: gradient matrix over hyperparameters. Shape: (n, m, l).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(cov_X_Xp, np.ndarray)
    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(num_hyps, int)
    assert isinstance(fix_noise, bool)

    num_X = X.shape[0]
    num_Xp = Xp.shape[0]

    grad_cov_X_Xp = np.zeros((num_X, num_Xp, num_hyps))
    dist = scisd.cdist(X / hyps['lengthscales'], Xp / hyps['lengthscales'], metric='euclidean')

    if fix_noise:
        ind_next = 0
    else:
        grad_cov_X_Xp[:, :, 0] += 2.0 * hyps['noise'] * np.eye(num_X, M=num_Xp)
        ind_next = 1

    grad_cov_X_Xp[:, :, ind_next] += 2.0 * cov_X_Xp / hyps['signal']

    if isinstance(hyps['lengthscales'], np.ndarray) and len(hyps['lengthscales'].shape) == 1:
        for ind_ in range(0, hyps['lengthscales'].shape[0]):
            grad_cov_X_Xp[:, :, ind_next+ind_+1] += cov_X_Xp * scisd.cdist(X[:, ind_][..., np.newaxis], Xp[:, ind_][..., np.newaxis], metric='euclidean')**2 * hyps['lengthscales'][ind_]**(-3)
    else:
        grad_cov_X_Xp[:, :, ind_next+1] += cov_X_Xp * dist**2 * hyps['lengthscales']**(-1)

    return grad_cov_X_Xp

@utils_common.validate_types
def cov_matern32(X: np.ndarray, Xp: np.ndarray, lengthscales: constants.TYPING_UNION_ARRAY_FLOAT,
    signal: float
) -> np.ndarray:
    """
    It computes Matern 3/2 kernel over `X` and `Xp`, where `lengthscales` and `signal` are given.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param lengthscales: length scales. Shape: (d, ) or ().
    :type lengthscales: numpy.ndarray, or float
    :param signal: coefficient for signal.
    :type signal: float

    :returns: kernel values over `X` and `Xp`. Shape: (n, m).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(lengthscales, (np.ndarray, float))
    assert len(X.shape) == 2
    assert len(Xp.shape) == 2
    if isinstance(lengthscales, np.ndarray):
        assert X.shape[1] == Xp.shape[1] == lengthscales.shape[0]
    else:
        assert X.shape[1] == Xp.shape[1]
    assert isinstance(signal, float)

    dist = scisd.cdist(X / lengthscales, Xp / lengthscales, metric='euclidean')
    cov_ = signal**2 * (1.0 + np.sqrt(3.0) * dist) * np.exp(-1.0 * np.sqrt(3.0) * dist)
    return cov_

@utils_common.validate_types
def grad_cov_matern32(cov_X_Xp: np.ndarray, X: np.ndarray, Xp: np.ndarray, hyps: dict,
    num_hyps: int, fix_noise: bool
) -> np.ndarray:
    """
    It computes gradients of Matern 3/2 kernel over `X` and `Xp`, where `hyps` is given.

    :param cov_X_Xp: covariance matrix. Shape: (n, m).
    :type cov_X_Xp: numpy.ndarray
    :param X: one inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param num_hyps: the number of hyperparameters == l.
    :type num_hyps: int.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool.

    :returns: gradient matrix over hyperparameters. Shape: (n, m, l).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(cov_X_Xp, np.ndarray)
    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(num_hyps, int)
    assert isinstance(fix_noise, bool)

    num_X = X.shape[0]
    num_Xp = Xp.shape[0]

    grad_cov_X_Xp = np.zeros((num_X, num_Xp, num_hyps))
    dist = scisd.cdist(X / hyps['lengthscales'], Xp / hyps['lengthscales'], metric='euclidean')

    if fix_noise:
        ind_next = 0
    else:
        grad_cov_X_Xp[:, :, 0] += 2.0 * hyps['noise'] * np.eye(num_X, M=num_Xp)
        ind_next = 1

    grad_cov_X_Xp[:, :, ind_next] += 2.0 * cov_X_Xp / hyps['signal']

    term_pre = 3.0 * hyps['signal']**2 * np.exp(-np.sqrt(3) * dist)

    if isinstance(hyps['lengthscales'], np.ndarray) and len(hyps['lengthscales'].shape) == 1:
        for ind_ in range(0, hyps['lengthscales'].shape[0]):
            grad_cov_X_Xp[:, :, ind_next+ind_+1] += term_pre * scisd.cdist(X[:, ind_][..., np.newaxis], Xp[:, ind_][..., np.newaxis], metric='euclidean')**2 * hyps['lengthscales'][ind_]**(-3)
    else:
        grad_cov_X_Xp[:, :, ind_next+1] += term_pre * dist**2 * hyps['lengthscales']**(-1)

    return grad_cov_X_Xp

@utils_common.validate_types
def cov_matern52(X: np.ndarray, Xp:np.ndarray, lengthscales: constants.TYPING_UNION_ARRAY_FLOAT,
    signal: float
) -> np.ndarray:
    """
    It computes Matern 5/2 kernel over `X` and `Xp`, where `lengthscales`
    and `signal` are given.

    :param X: inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param lengthscales: length scales. Shape: (d, ) or ().
    :type lengthscales: numpy.ndarray, or float
    :param signal: coefficient for signal.
    :type signal: float

    :returns: kernel values over `X` and `Xp`. Shape: (n, m).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(lengthscales, (np.ndarray, float))
    assert len(X.shape) == 2
    assert len(Xp.shape) == 2
    if isinstance(lengthscales, np.ndarray):
        assert X.shape[1] == Xp.shape[1] == lengthscales.shape[0]
    else:
        assert X.shape[1] == Xp.shape[1]
    assert isinstance(signal, float)

    dist = scisd.cdist(X / lengthscales, Xp / lengthscales, metric='euclidean')
    cov_X_Xp = signal**2 * (1.0 + np.sqrt(5.0) * dist + 5.0 / 3.0 * dist**2) \
        * np.exp(-1.0 * np.sqrt(5.0) * dist)
    return cov_X_Xp

@utils_common.validate_types
def grad_cov_matern52(cov_X_Xp: np.ndarray, X: np.ndarray, Xp: np.ndarray, hyps: dict,
    num_hyps: int, fix_noise: bool
) -> np.ndarray:
    """
    It computes gradients of Matern 5/2 kernel over `X` and `Xp`, where `hyps` is given.

    :param cov_X_Xp: covariance matrix. Shape: (n, m).
    :type cov_X_Xp: numpy.ndarray
    :param X: one inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param num_hyps: the number of hyperparameters == l.
    :type num_hyps: int.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool.

    :returns: gradient matrix over hyperparameters. Shape: (n, m, l).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(cov_X_Xp, np.ndarray)
    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(num_hyps, int)
    assert isinstance(fix_noise, bool)

    num_X = X.shape[0]
    num_Xp = Xp.shape[0]

    grad_cov_X_Xp = np.zeros((num_X, num_Xp, num_hyps))
    dist = scisd.cdist(X / hyps['lengthscales'], Xp / hyps['lengthscales'], metric='euclidean')

    if fix_noise:
        ind_next = 0
    else:
        grad_cov_X_Xp[:, :, 0] += 2.0 * hyps['noise'] * np.eye(num_X, M=num_Xp)
        ind_next = 1

    grad_cov_X_Xp[:, :, ind_next] += 2.0 * cov_X_Xp / hyps['signal']

    term_pre = 5.0 / 3.0 * hyps['signal']**2 * (1.0 + np.sqrt(5) * dist) \
        * np.exp(-np.sqrt(5) * dist)

    if isinstance(hyps['lengthscales'], np.ndarray) and len(hyps['lengthscales'].shape) == 1:
        for ind_ in range(0, hyps['lengthscales'].shape[0]):
            grad_cov_X_Xp[:, :, ind_next+ind_+1] += term_pre * scisd.cdist(X[:, ind_][..., np.newaxis], Xp[:, ind_][..., np.newaxis], metric='euclidean')**2 * hyps['lengthscales'][ind_]**(-3)
    else:
        grad_cov_X_Xp[:, :, ind_next+1] += term_pre * hyps['lengthscales']**(-1) * dist**2

    return grad_cov_X_Xp

@utils_common.validate_types
def cov_set(str_cov: str, X: np.ndarray, Xp: np.ndarray,
    lengthscales: constants.TYPING_UNION_ARRAY_FLOAT, signal: float
) -> np.ndarray:
    """
    It computes set kernel matrix over `X` and `Xp`, where `lengthscales` and `signal` are given.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param X: one inputs. Shape: (n, m, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (l, m, d).
    :type Xp: numpy.ndarray
    :param lengthscales: length scales. Shape: (d, ) or ().
    :type lengthscales: numpy.ndarray, or float
    :param signal: coefficient for signal.
    :type signal: float

    :returns: set kernel matrix over `X` and `Xp`. Shape: (n, l).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)
    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(lengthscales, (np.ndarray, float))
    assert isinstance(signal, float)
    assert len(X.shape) == 2
    assert len(Xp.shape) == 2
    if isinstance(lengthscales, np.ndarray):
        assert X.shape[1] == Xp.shape[1] == lengthscales.shape[0]
    else:
        assert X.shape[1] == Xp.shape[1]
    assert str_cov in constants.ALLOWED_COV_BASE
    num_X = X.shape[0]
    num_Xp = Xp.shape[0]

    fun_cov = choose_fun_cov(str_cov)
    cov_X_Xp = fun_cov(X, Xp, lengthscales, signal)
    cov_X_Xp = np.sum(cov_X_Xp)

    cov_X_Xp /= num_X * num_Xp

    return cov_X_Xp

@utils_common.validate_types
def cov_main(str_cov: str, X: np.ndarray, Xp: np.ndarray, hyps: dict, same_X_Xp: bool,
    jitter: float=constants.JITTER_COV
) -> np.ndarray:
    """
    It computes kernel matrix over `X` and `Xp`, where `hyps` is given.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param X: one inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param same_X_Xp: flag for checking `X` and `Xp` are same.
    :type same_X_Xp: bool.
    :param jitter: jitter for diagonal entries.
    :type jitter: float, optional

    :returns: kernel matrix over `X` and `Xp`. Shape: (n, m).
    :rtype: numpy.ndarray

    :raises: AssertionError, ValueError

    """

    assert isinstance(str_cov, str)
    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(same_X_Xp, bool)
    assert isinstance(jitter, float)
    assert str_cov in constants.ALLOWED_COV

    num_X = X.shape[0]
    num_Xp = Xp.shape[0]

    cov_X_Xp = np.zeros((num_X, num_Xp))
    if same_X_Xp:
        assert num_X == num_Xp
        cov_X_Xp += np.eye(num_X) * jitter

    if str_cov in constants.ALLOWED_COV_BASE:
        assert len(X.shape) == 2
        assert len(Xp.shape) == 2
        dim_X = X.shape[1]
        dim_Xp = Xp.shape[1]
        assert dim_X == dim_Xp

        hyps = utils_covariance.validate_hyps_dict(hyps, str_cov, dim_X)

        fun_cov = choose_fun_cov(str_cov)
        cov_X_Xp += fun_cov(X, Xp, hyps['lengthscales'], hyps['signal'])

        assert cov_X_Xp.shape == (num_X, num_Xp)
    elif str_cov in constants.ALLOWED_COV_SET:
        list_str_cov = str_cov.split('_')
        str_cov = list_str_cov[1]

        assert len(X.shape) == 3
        assert len(Xp.shape) == 3

        dim_X = X.shape[2]
        dim_Xp = Xp.shape[2]

        assert dim_X == dim_Xp

        hyps = utils_covariance.validate_hyps_dict(hyps, str_cov, dim_X)

        if not same_X_Xp:
            for ind_X in range(0, num_X):
                for ind_Xp in range(0, num_Xp):
                    cov_X_Xp[ind_X, ind_Xp] += cov_set(str_cov, X[ind_X], Xp[ind_Xp],
                        hyps['lengthscales'], hyps['signal'])
        else:
            for ind_X in range(0, num_X):
                for ind_Xp in range(ind_X, num_Xp):
                    cov_X_Xp[ind_X, ind_Xp] += cov_set(str_cov, X[ind_X], Xp[ind_Xp],
                        hyps['lengthscales'], hyps['signal'])
                    if ind_X < ind_Xp:
                        cov_X_Xp[ind_Xp, ind_X] = cov_X_Xp[ind_X, ind_Xp]
    else:
        raise NotImplementedError('cov_main: allowed str_cov, but it is not implemented.')

    return cov_X_Xp

@utils_common.validate_types
def grad_cov_main(str_cov: str, X: np.ndarray, Xp: np.ndarray, hyps: dict, fix_noise: bool,
    same_X_Xp: bool=True,
    jitter: float=constants.JITTER_COV,
) -> np.ndarray:
    """
    It computes gradients of kernel matrix over hyperparameters, where `hyps` is given.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param X: one inputs. Shape: (n, d).
    :type X: numpy.ndarray
    :param Xp: another inputs. Shape: (m, d).
    :type Xp: numpy.ndarray
    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool.
    :param same_X_Xp: flag for checking `X` and `Xp` are same.
    :type same_X_Xp: bool., optional
    :param jitter: jitter for diagonal entries.
    :type jitter: float, optional

    :returns: gradient matrix over hyperparameters. Shape: (n, m, l) where
        l is the number of hyperparameters.
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)
    assert isinstance(X, np.ndarray)
    assert isinstance(Xp, np.ndarray)
    assert isinstance(hyps, dict)
    assert isinstance(fix_noise, bool)
    assert isinstance(same_X_Xp, bool)
    assert isinstance(jitter, float)
    assert str_cov in constants.ALLOWED_COV
    # TODO: X and Xp should be same?
    assert same_X_Xp

    dim_X = X.shape[1]

    if isinstance(hyps['lengthscales'], np.ndarray):
        num_hyps = dim_X + 1
    else:
        num_hyps = 2

    if not fix_noise:
        num_hyps += 1

    cov_X_Xp = cov_main(str_cov, X, Xp, hyps, same_X_Xp, jitter=jitter)

    fun_grad_cov = choose_fun_grad_cov(str_cov)
    grad_cov_X_Xp = fun_grad_cov(cov_X_Xp, X, Xp, hyps, num_hyps, fix_noise)

    return grad_cov_X_Xp



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: December 29, 2020
#
"""These files are for implementing Student-:math:`t` process regression.
It is implemented, based on the following article:

(i) Rasmussen, C. E., & Williams, C. K. (2006). Gaussian Process
Regression for Machine Learning. MIT Press.

(ii) Shah, A., Wilson, A. G., & Ghahramani, Z. (2014). Student-t Processes
as Alternatives to Gaussian Processes. In Proceedings of the 17th
International Conference on Artificial Intelligence and Statistics
(pp. 877-885)."""



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: December 29, 2020
#
"""It is utilities for covariance functions."""

import numpy as np

from bayeso.utils import utils_common
from bayeso import constants


@utils_common.validate_types
def _get_list_first() -> constants.TYPING_LIST[str]:
    """
    It provides list of strings.
    The strings in that list require two hyperparameters, `signal` and `lengthscales`.
    We simply call it as `list_first`.

    :returns: list of strings, which satisfy some requirements we mentioned above.
    :rtype: list

    """

    list_first = ['eq', 'se', 'matern32', 'matern52']
    list_first += ['set_' + str_ for str_ in list_first]
    return list_first

@utils_common.validate_types
def get_hyps(str_cov: str, dim: int,
    use_gp: bool=True,
    use_ard: bool=True,
) -> dict:
    """
    It returns a dictionary of default hyperparameters for covariance
    function, where `str_cov` and `dim` are given. If `use_ard` is True,
    the length scales would be `dim`-dimensional vector.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param dim: dimensionality of the problem we are solving.
    :type dim: int.
    :param use_gp: flag for Gaussian process or Student-$t$ process.
    :type use_gp: bool., optional
    :param use_ard: flag for automatic relevance determination.
    :type use_ard: bool., optional

    :returns: dictionary of default hyperparameters for covariance function.
    :rtype: dict.

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)
    assert isinstance(dim, int)
    assert isinstance(use_gp, bool)
    assert isinstance(use_ard, bool)
    assert str_cov in constants.ALLOWED_COV

    hyps = dict()
    hyps['noise'] = constants.GP_NOISE

    if not use_gp:
        hyps['dof'] = 5.0

    list_first = _get_list_first()

    if str_cov in list_first:
        hyps['signal'] = 1.0
        if use_ard:
            hyps['lengthscales'] = np.ones(dim)
        else:
            hyps['lengthscales'] = 1.0
    else:
        raise NotImplementedError('get_hyps: allowed str_cov, but it is not implemented.')
    return hyps

@utils_common.validate_types
def get_range_hyps(str_cov: str, dim: int,
    use_gp: bool=True,
    use_ard: bool=True,
    fix_noise: bool=False
) -> constants.TYPING_LIST[list]:
    """
    It returns default optimization ranges of hyperparameters for Gaussian process regression.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param dim: dimensionality of the problem we are solving.
    :type dim: int.
    :param use_gp: flag for Gaussian process or Student-$t$ process.
    :type use_gp: bool., optional
    :param use_ard: flag for automatic relevance determination.
    :type use_ard: bool., optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional

    :returns: list of default optimization ranges for hyperparameters.
    :rtype: list

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)
    assert isinstance(dim, int)
    assert isinstance(use_gp, bool)
    assert isinstance(use_ard, bool)
    assert isinstance(fix_noise, bool)
    assert str_cov in constants.ALLOWED_COV

    range_hyps = []

    list_first = _get_list_first()

    if not fix_noise:
        range_hyps += constants.RANGE_NOISE

    if not use_gp:
        range_hyps += constants.RANGE_DOF

    if str_cov in list_first:
        range_hyps += constants.RANGE_SIGNAL # for signal scale
        if use_ard: # for lengthscales
            for _ in range(0, dim):
                range_hyps += constants.RANGE_LENGTHSCALES
        else:
            # INFO: dim is ignored.
            range_hyps += constants.RANGE_LENGTHSCALES
    else:
        raise NotImplementedError('get_hyps: allowed str_cov, but it is not implemented.')

    return range_hyps

@utils_common.validate_types
def convert_hyps(str_cov: str, hyps: dict,
    use_gp: bool=True,
    fix_noise: bool=False
) -> np.ndarray:
    """
    It converts hyperparameters dictionary, `hyps` to numpy array.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param use_gp: flag for Gaussian process or Student-$t$ process.
    :type use_gp: bool., optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional

    :returns: converted array of the hyperparameters given by `hyps`.
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)
    assert isinstance(hyps, dict)
    assert isinstance(use_gp, bool)
    assert isinstance(fix_noise, bool)
    assert str_cov in constants.ALLOWED_COV

    list_hyps = []
    if not fix_noise:
        list_hyps.append(hyps['noise'])

    if not use_gp:
        list_hyps.append(hyps['dof'])

    list_first = _get_list_first()

    if str_cov in list_first:
        list_hyps.append(hyps['signal'])
        if isinstance(hyps['lengthscales'], np.ndarray):
            for elem_lengthscale in hyps['lengthscales']:
                list_hyps.append(elem_lengthscale)
        elif isinstance(hyps['lengthscales'], float):
            list_hyps.append(hyps['lengthscales'])
        else: # pragma: no cover
            raise ValueError('covert_hyps: not allowed type for lengthscales.')
    else:
        raise NotImplementedError('convert_hyps: allowed str_cov, but it is not implemented.')

    return np.array(list_hyps)

@utils_common.validate_types
def restore_hyps(str_cov: str, hyps: np.ndarray,
    use_gp: bool=True,
    use_ard: bool=True,
    fix_noise: bool=False,
    noise: float=constants.GP_NOISE
) -> dict:
    """
    It restores hyperparameters array, `hyps` to dictionary.

    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param hyps: array of hyperparameters for covariance function.
    :type hyps: numpy.ndarray
    :param use_gp: flag for Gaussian process or Student-$t$ process.
    :type use_gp: bool., optional
    :param use_ard: flag for using automatic relevance determination.
    :type use_ard: bool., optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param noise: fixed noise value.
    :type noise: float, optional

    :returns: restored dictionary of the hyperparameters given by `hyps`.
    :rtype: dict.

    :raises: AssertionError

    """

    assert isinstance(str_cov, str)
    assert isinstance(hyps, np.ndarray)
    assert isinstance(use_gp, bool)
    assert isinstance(use_ard, bool)
    assert isinstance(fix_noise, bool)
    assert isinstance(noise, float)
    assert len(hyps.shape) == 1
    assert str_cov in constants.ALLOWED_COV

    dict_hyps = dict()
    if not fix_noise:
        dict_hyps['noise'] = hyps[0]
        ind_start = 1
    else:
        dict_hyps['noise'] = noise
        ind_start = 0

    if not use_gp:
        dict_hyps['dof'] = hyps[ind_start]
        ind_start += 1

    list_first = _get_list_first()

    if str_cov in list_first:
        dict_hyps['signal'] = hyps[ind_start]

        if use_ard:
            list_lengthscales = []
            for ind_elem in range(ind_start + 1, len(hyps)):
                list_lengthscales.append(hyps[ind_elem])
            dict_hyps['lengthscales'] = np.array(list_lengthscales)
        else:
            assert hyps.shape[0] == ind_start + 2
            dict_hyps['lengthscales'] = hyps[ind_start + 1]
    else:
        raise NotImplementedError('restore_hyps: allowed str_cov, but it is not implemented.')
    return dict_hyps

@utils_common.validate_types
def validate_hyps_dict(hyps: dict, str_cov: str, dim: int,
    use_gp: bool=True
) -> constants.TYPING_TUPLE_DICT_BOOL:
    """
    It validates hyperparameters dictionary, `hyps`.

    :param hyps: dictionary of hyperparameters for covariance function.
    :type hyps: dict.
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param dim: dimensionality of the problem we are solving.
    :type dim: int.
    :param use_gp: flag for Gaussian process or Student-$t$ process.
    :type use_gp: bool., optional

    :returns: a tuple of valid hyperparameters and validity flag.
    :rtype: (dict., bool.)

    :raises: AssertionError

    """

    assert isinstance(hyps, dict)
    assert isinstance(str_cov, str)
    assert isinstance(dim, int)
    assert isinstance(use_gp, bool)
    assert str_cov in constants.ALLOWED_COV

    if 'noise' not in hyps:
        raise ValueError('validate_hyps_dict: invalid noise.')

    if not isinstance(hyps['noise'], float):
        raise ValueError('validate_hyps_dict: invalid noise.')

    if np.abs(hyps['noise']) >= constants.BOUND_UPPER_GP_NOISE:
        hyps['noise'] = constants.BOUND_UPPER_GP_NOISE

    if not use_gp:
        if 'dof' not in hyps:
            raise ValueError('validate_hyps_dict: invalid dof.')

        if not isinstance(hyps['dof'], float):
            raise ValueError('validate_hyps_dict: invalid dof.')

        if isinstance(hyps['dof'], float) and hyps['dof'] <= 2.0:
            hyps['dof'] = 2.00001

    if 'lengthscales' not in hyps:
        raise ValueError('validate_hyps_dict: invalid lengthscales.')

    if isinstance(hyps['lengthscales'], np.ndarray) \
        and hyps['lengthscales'].shape[0] != dim:
        raise ValueError('validate_hyps_dict: invalid lengthscales.')
    if not isinstance(hyps['lengthscales'], np.ndarray) \
        and not isinstance(hyps['lengthscales'], float):
        raise ValueError('validate_hyps_dict: invalid lengthscales.')

    if 'signal' not in hyps:
        raise ValueError('validate_hyps_dict: invalid signal.')

    if not isinstance(hyps['signal'], float):
        raise ValueError('validate_hyps_dict: invalid signal.')

    return hyps

@utils_common.validate_types
def validate_hyps_arr(hyps: np.ndarray, str_cov: str, dim: int,
    use_gp: bool=True
) -> constants.TYPING_TUPLE_ARRAY_BOOL:
    """
    It validates hyperparameters array, `hyps`.

    :param hyps: array of hyperparameters for covariance function.
    :type hyps: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param dim: dimensionality of the problem we are solving.
    :type dim: int.
    :param use_gp: flag for Gaussian process or Student-$t$ process.
    :type use_gp: bool., optional

    :returns: a tuple of valid hyperparameters and validity flag.
    :rtype: (numpy.ndarray, bool.)

    :raises: AssertionError

    """

    assert isinstance(hyps, np.ndarray)
    assert isinstance(str_cov, str)
    assert isinstance(dim, int)
    assert isinstance(use_gp, bool)
    assert str_cov in constants.ALLOWED_COV

#    is_valid = True

    raise NotImplementedError('validate_hyps_arr in utils_covariance.py')

@utils_common.validate_types
def check_str_cov(str_fun: str, str_cov: str, shape_X1: tuple,
    shape_X2: tuple=None
) -> constants.TYPE_NONE:
    """
    It is for validating the shape of X1 (and optionally the shape of X2).

    :param str_fun: the name of function.
    :type str_fun: str.
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param shape_X1: the shape of X1.
    :type shape_X1: tuple
    :param shape_X2: None, or the shape of X2.
    :type shape_X2: NoneType or tuple, optional

    :returns: None, if it is valid. Raise an error, otherwise.
    :rtype: NoneType

    :raises: AssertionError, ValueError

    """

    assert isinstance(str_fun, str)
    assert isinstance(str_cov, str)
    assert isinstance(shape_X1, tuple)
    assert shape_X2 is None or isinstance(shape_X2, tuple)

    if str_cov in constants.ALLOWED_COV_BASE:
        assert len(shape_X1) == 2
        if shape_X2 is not None:
            assert len(shape_X2) == 2
    elif str_cov in constants.ALLOWED_COV_SET:
        assert len(shape_X1) == 3
        if shape_X2 is not None:
            assert len(shape_X2) == 3
    elif str_cov in constants.ALLOWED_COV: # pragma: no cover
        raise ValueError('{}: missing conditions for str_cov.'.format(str_fun))
    else:
        raise ValueError('{}: invalid str_cov.'.format(str_fun))



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: March 22, 2021
#
"""It defines the functions related to likelihood for
Student-:math:`t` process regression."""

import numpy as np
import scipy.special

from bayeso import covariance
from bayeso import constants
from bayeso.utils import utils_gp
from bayeso.utils import utils_covariance
from bayeso.utils import utils_common
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('tp_likelihood')


@utils_common.validate_types
def neg_log_ml(X_train: np.ndarray, Y_train: np.ndarray, hyps: np.ndarray,
    str_cov: str, prior_mu_train: np.ndarray,
    use_ard: bool=constants.USE_ARD,
    fix_noise: bool=constants.FIX_GP_NOISE,
    use_gradient: bool=True,
    debug: bool=False
) -> constants.TYPING_UNION_FLOAT_FA:
    """
    This function computes a negative log marginal likelihood.

    :param X_train: inputs. Shape: (n, d) or (n, m, d).
    :type X_train: numpy.ndarray
    :param Y_train: outputs. Shape: (n, 1).
    :type Y_train: numpy.ndarray
    :param hyps: hyperparameters for Gaussian process. Shape: (h, ).
    :type hyps: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str.
    :param prior_mu_train: the prior values computed by get_prior_mu(). Shape: (n, 1).
    :type prior_mu_train: numpy.ndarray
    :param use_ard: flag for automatic relevance determination.
    :type use_ard: bool., optional
    :param fix_noise: flag for fixing a noise.
    :type fix_noise: bool., optional
    :param use_gradient: flag for computing and returning gradients of
        negative log marginal likelihood.
    :type use_gradient: bool., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    :returns: negative log marginal likelihood, or (negative log marginal
        likelihood, gradients of the likelihood).
    :rtype: float, or tuple of (float, np.ndarray)

    :raises: AssertionError

    """

    utils_gp.validate_common_args(X_train, Y_train, str_cov, None, debug)
    assert isinstance(hyps, np.ndarray)
    assert isinstance(prior_mu_train, np.ndarray)
    assert isinstance(use_ard, bool)
    assert isinstance(fix_noise, bool)
    assert isinstance(use_gradient, bool)
    assert len(prior_mu_train.shape) == 2
    assert X_train.shape[0] == Y_train.shape[0] == prior_mu_train.shape[0]
    utils_covariance.check_str_cov('neg_log_ml', str_cov, X_train.shape)

    num_X = float(X_train.shape[0])
    hyps = utils_covariance.restore_hyps(str_cov, hyps,
        use_ard=use_ard,
        fix_noise=fix_noise, use_gp=False)
    new_Y_train = Y_train - prior_mu_train
    nu = hyps['dof']

    cov_X_X, inv_cov_X_X, grad_cov_X_X = covariance.get_kernel_inverse(X_train,
        hyps, str_cov, fix_noise=fix_noise, use_gradient=use_gradient,
        debug=debug)

    alpha = np.dot(inv_cov_X_X, new_Y_train)
    beta = np.squeeze(np.dot(np.dot(new_Y_train.T, inv_cov_X_X), new_Y_train))

    first_term = -0.5 * num_X * np.log((nu - 2.0) * np.pi)
    sign_second_term, second_term = np.linalg.slogdet(cov_X_X)

    # TODO: it should be checked.
    if sign_second_term <= 0: # pragma: no cover
        second_term = 0.0

    second_term = -0.5 * second_term

    third_term = np.log(scipy.special.gamma((nu + num_X) / 2.0) / scipy.special.gamma(nu / 2.0))
    fourth_term = -0.5 * (nu + num_X) * np.log(1.0 + beta / (nu - 2.0))

    log_ml_ = np.squeeze(first_term + second_term + third_term + fourth_term)
    log_ml_ /= num_X

    if use_gradient:
        assert grad_cov_X_X is not None
        grad_log_ml_ = np.zeros(grad_cov_X_X.shape[2] + 1)

        first_term_grad = ((nu + num_X) / (nu + beta - 2.0) * np.dot(alpha, alpha.T) - inv_cov_X_X)
        nu_grad = -num_X / (2.0 * (nu - 2.0))\
            + scipy.special.digamma((nu + num_X) / 2.0)\
            - scipy.special.digamma(nu / 2.0)\
            - 0.5 * np.log(1.0 + beta / (nu - 2.0))\
            + (nu + num_X) * beta / (2.0 * (nu - 2.0)**2 + 2.0 * beta * (nu - 2.0))

        if fix_noise:
            grad_log_ml_[0] = nu_grad
        else:
            grad_log_ml_[1] = nu_grad

        for ind in range(0, grad_cov_X_X.shape[2]):
            cur_grad = 0.5 * np.trace(np.dot(first_term_grad, grad_cov_X_X[:, :, ind]))
            if fix_noise:
                grad_log_ml_[ind + 1] = cur_grad
            else:
                if ind == 0:
                    cur_ind = 0
                else:
                    cur_ind = ind + 1

                grad_log_ml_[cur_ind] = cur_grad

    if use_gradient:
        return -1.0 * log_ml_, -1.0 * grad_log_ml_ / num_X

    return -1.0 * log_ml_



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: July 8, 2021
#
"""It defines a wrapper class for Bayesian optimization."""

import time
import numpy as np

from bayeso import constants
from bayeso import bo
from bayeso.utils import utils_bo
from bayeso.utils import utils_logger

logger = utils_logger.get_logger('wrappers_bayesian_optimization')


class BayesianOptimization:
    """
    It is a wrapper class for Bayesian optimization.
    A function for optimizing `fun_target` runs a single round
    of Bayesian optimization with an iteration budget `num_iter`.

    :param range_X: a search space. Shape: (d, 2).
    :type range_X: numpy.ndarray
    :param fun_target: a target function.
    :type fun_target: callable
    :param num_iter: an iteration budget for Bayesian optimization.
    :type num_iter: int.
    :param str_surrogate: the name of surrogate model.
    :type str_surrogate: str., optional
    :param str_cov: the name of covariance function.
    :type str_cov: str., optional
    :param str_acq: the name of acquisition function.
    :type str_acq: str., optional
    :param normalize_Y: a flag for normalizing outputs.
    :type normalize_Y: bool., optional
    :param use_ard: a flag for automatic relevance determination.
    :type use_ard: bool., optional
    :param prior_mu: None, or a prior mean function.
    :type prior_mu: NoneType, or callable, optional
    :param str_initial_method_bo: the name of initialization method for
        sampling initial examples in Bayesian optimization.
    :type str_initial_method_bo: str., optional
    :param str_sampling_method_ao: the name of sampling method for
        acquisition function optimization.
    :type str_sampling_method_ao: str., optional
    :param str_optimizer_method_gp: the name of optimization method for
        Gaussian process regression.
    :type str_optimizer_method_gp: str., optional
    :param str_optimizer_method_bo: the name of optimization method for
        Bayesian optimization.
    :type str_optimizer_method_bo: str., optional
    :param str_mlm_method: the name of marginal likelihood maximization
        method for Gaussian process regression.
    :type str_mlm_method: str., optional
    :param str_modelselection_method: the name of model selection method
        for Gaussian process regression.
    :type str_modelselection_method: str., optional
    :param num_samples_ao: the number of samples for acquisition function
        optimization. If a local search method (e.g., L-BFGS-B) is selected
        for acquisition function optimization, it is employed.
    :type num_samples_ao: int., optional
    :param debug: a flag for printing log messages.
    :type debug: bool., optional

    """

    def __init__(self,
        range_X: np.ndarray,
        fun_target: constants.TYPING_CALLABLE,
        num_iter: int,
        str_surrogate: str=constants.STR_SURROGATE,
        str_cov: str=constants.STR_COV,
        str_acq: str=constants.STR_BO_ACQ,
        normalize_Y: bool=constants.NORMALIZE_RESPONSE,
        use_ard: bool=constants.USE_ARD,
        prior_mu: constants.TYPING_UNION_CALLABLE_NONE=None,
        str_initial_method_bo: str=constants.STR_INITIALIZING_METHOD_BO,
        str_sampling_method_ao: str=constants.STR_SAMPLING_METHOD_AO,
        str_optimizer_method_gp: str=constants.STR_OPTIMIZER_METHOD_GP,
        str_optimizer_method_tp: str=constants.STR_OPTIMIZER_METHOD_TP,
        str_optimizer_method_bo: str=constants.STR_OPTIMIZER_METHOD_AO,
        str_mlm_method: str=constants.STR_MLM_METHOD,
        str_modelselection_method: str=constants.STR_MODELSELECTION_METHOD,
        num_samples_ao: int=constants.NUM_SAMPLES_AO,
        debug: bool=False,
    ):
        """
        Constructor method

        """

        assert isinstance(range_X, np.ndarray)
        assert callable(fun_target)
        assert isinstance(num_iter, int)
        assert isinstance(str_surrogate, str)
        assert isinstance(str_cov, str)
        assert isinstance(str_acq, str)
        assert isinstance(normalize_Y, bool)
        assert isinstance(use_ard, bool)
        assert callable(prior_mu) or prior_mu is None
        assert isinstance(str_initial_method_bo, str)
        assert isinstance(str_sampling_method_ao, str)
        assert isinstance(str_optimizer_method_gp, str)
        assert isinstance(str_optimizer_method_tp, str)
        assert isinstance(str_optimizer_method_bo, str)
        assert isinstance(str_mlm_method, str)
        assert isinstance(str_modelselection_method, str)
        assert isinstance(num_samples_ao, int)
        assert isinstance(debug, bool)

        assert len(range_X.shape) == 2
        assert range_X.shape[1] == 2
        assert (range_X[:, 0] <= range_X[:, 1]).all()
        assert num_iter > 0
        assert num_samples_ao > 0

        assert str_surrogate in constants.ALLOWED_SURROGATE + constants.ALLOWED_SURROGATE_TREES
        assert str_cov in constants.ALLOWED_COV
        assert str_acq in constants.ALLOWED_BO_ACQ
        assert str_initial_method_bo in constants.ALLOWED_INITIALIZING_METHOD_BO
        assert str_sampling_method_ao in constants.ALLOWED_SAMPLING_METHOD
        assert str_optimizer_method_gp in constants.ALLOWED_OPTIMIZER_METHOD_GP
        assert str_optimizer_method_tp in constants.ALLOWED_OPTIMIZER_METHOD_TP
        assert str_optimizer_method_bo in constants.ALLOWED_OPTIMIZER_METHOD_BO \
            + constants.ALLOWED_OPTIMIZER_METHOD_BO_TREES
        assert str_mlm_method in constants.ALLOWED_MLM_METHOD
        assert str_modelselection_method in constants.ALLOWED_MODELSELECTION_METHOD

        self.range_X = range_X
        self.num_dim = range_X.shape[0]
        self.fun_target = fun_target
        self.num_iter = num_iter
        self.str_surrogate = str_surrogate
        self.str_acq = str_acq
        self.normalize_Y = normalize_Y
        self.str_initial_method_bo = str_initial_method_bo
        self.str_sampling_method_ao = str_sampling_method_ao
        self.str_optimizer_method_bo = utils_bo.check_optimizer_method_bo(
            str_optimizer_method_bo, range_X.shape[0], debug)
        self.num_samples_ao = num_samples_ao
        self.debug = debug

        if str_surrogate in constants.ALLOWED_SURROGATE_TREES:
            self.model_bo = self._get_model_bo_trees()
        elif str_surrogate == 'gp':
            self.str_cov = str_cov
            self.use_ard = use_ard
            self.prior_mu = prior_mu
            self.str_optimizer_method_gp = str_optimizer_method_gp
            self.str_mlm_method = str_mlm_method
            self.str_modelselection_method = str_modelselection_method

            self.model_bo = self._get_model_bo_gp()
        elif str_surrogate == 'tp':
            self.str_cov = str_cov
            self.use_ard = use_ard
            self.prior_mu = prior_mu
            self.str_optimizer_method_tp = str_optimizer_method_tp

            self.model_bo = self._get_model_bo_tp()
        else:
            raise NotImplementedError('allowed str_surrogate, but it is not implemented.')

    def _get_model_bo_gp(self):
        """
        It returns an object of `bayeso.bo.bo_w_gp.BO_w_GP`.

        :returns: an object of Bayesian optimization.
        :rtype: bayeso.bo.bo_w_gp.BOwGP

        """

        model_bo = bo.BOwGP(
            self.range_X,
            str_cov=self.str_cov,
            str_acq=self.str_acq,
            normalize_Y=self.normalize_Y,
            use_ard=self.use_ard,
            prior_mu=self.prior_mu,
            str_optimizer_method_gp=self.str_optimizer_method_gp,
            str_optimizer_method_bo=self.str_optimizer_method_bo,
            str_modelselection_method=self.str_modelselection_method,
            debug=self.debug
        )

        return model_bo

    def _get_model_bo_tp(self):
        """
        It returns an object of `bayeso.bo.bo_w_tp.BO_w_TP`.

        :returns: an object of Bayesian optimization.
        :rtype: bayeso.bo.bo_w_tp.BOwTP

        """

        model_bo = bo.BOwTP(
            self.range_X,
            str_cov=self.str_cov,
            str_acq=self.str_acq,
            normalize_Y=self.normalize_Y,
            use_ard=self.use_ard,
            prior_mu=self.prior_mu,
            str_optimizer_method_tp=self.str_optimizer_method_tp,
            str_optimizer_method_bo=self.str_optimizer_method_bo,
            debug=self.debug
        )

        return model_bo

    def _get_model_bo_trees(self):
        """
        It returns an object of `bayeso.bo.bo_w_trees.BO_w_Trees`.

        :returns: an object of Bayesian optimization.
        :rtype: bayeso.bo.bo_w_trees.BOwTrees

        """

        model_bo = bo.BOwTrees(
            self.range_X,
            str_surrogate=self.str_surrogate,
            str_acq=self.str_acq,
            normalize_Y=self.normalize_Y,
            str_optimizer_method_bo=self.str_optimizer_method_bo,
            debug=self.debug
        )

        return model_bo

    def _get_next_best_sample(self,
        next_sample: np.ndarray,
        X: np.ndarray,
        next_samples: np.ndarray,
        acq_vals: np.ndarray,
    ) -> np.ndarray:
        """
        It returns the next best sample in terms of acquisition function values.

        :param next_sample: the next sample acquired.
        :type next_sample: np.ndarray
        :param X: the samples evaluated so far.
        :type X: np.ndarray
        :param next_samples: the candidates of the next sample.
        :type next_samples: np.ndarray
        :param acq_vals: the values of acquisition function over `next_samples`.
        :type acq_vals: np.ndarray

        :returns: the next best sample. Shape: (d, ).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(next_sample, np.ndarray)
        assert isinstance(X, np.ndarray)
        assert isinstance(next_samples, np.ndarray)
        assert isinstance(acq_vals, np.ndarray)

        if np.where(np.linalg.norm(next_sample - X, axis=1)\
            < constants.TOLERANCE_DUPLICATED_ACQ)[0].shape[0] > 0: # pragma: no cover
            next_sample = utils_bo.get_next_best_acquisition(
                next_samples, acq_vals, X)

            if self.debug:
                logger.debug('next_sample is repeated, so next best is selected.\
                    next_sample: %s', utils_logger.get_str_array(next_sample))
        return next_sample

    def optimize_single_iteration(self, X: np.ndarray, Y: np.ndarray
    ) -> constants.TYPING_TUPLE_ARRAY_DICT:
        """
        It returns the optimization result and time consumed
        of single iteration, given `X` and `Y`.

        :param X: inputs. Shape: (n, d) or (n, m, d).
        :type X: numpy.ndarray
        :param Y: outputs. Shape: (n, 1).
        :type Y: numpy.ndarray

        :returns: a tuple of the next sample and information dictionary.
        :rtype: (numpy.ndarray, dict.)

        :raises: AssertionError, NotImplementedError

        """

        if self.str_surrogate in constants.ALLOWED_SURROGATE_TREES:
            next_sample, dict_info = self.model_bo.optimize(X, Y,
                str_sampling_method=self.str_sampling_method_ao,
                num_samples=self.num_samples_ao)
        elif self.str_surrogate == 'gp':
            next_sample, dict_info = self.model_bo.optimize(X, Y,
                str_sampling_method=self.str_sampling_method_ao,
                num_samples=self.num_samples_ao,
                str_mlm_method=self.str_mlm_method)
        elif self.str_surrogate == 'tp':
            next_sample, dict_info = self.model_bo.optimize(X, Y,
                str_sampling_method=self.str_sampling_method_ao,
                num_samples=self.num_samples_ao)
        else:
            raise NotImplementedError('allowed str_surrogate, but it is not implemented.')

        return next_sample, dict_info

    def optimize_with_all_initial_information(self,
        X: np.ndarray, Y: np.ndarray,
    ) -> constants.TYPING_TUPLE_FIVE_ARRAYS:
        """
        It returns the optimization results and times consumed, given
        inital inputs `X` and their corresponding outputs `Y`.

        :param X: initial inputs. Shape: (n, d) or (n, m, d).
        :type X: numpy.ndarray
        :param Y: initial outputs. Shape: (n, 1).
        :type Y: numpy.ndarray

        :returns: a tuple of acquired samples, their function values, overall
            times consumed per iteration, time consumed in modeling Gaussian process
            regression, and time consumed in acquisition function optimization.
            Shape: ((n + `num_iter`, d), (n + `num_iter`, 1),
            (`num_iter`, ), (`num_iter`, ), (`num_iter`, )),
            or ((n + `num_iter`, m, d),
            (n + `num_iter`, m, 1), (`num_iter`, ), (`num_iter`, ),
            (`num_iter`, )).
        :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)

        :raises: AssertionError

        """

        assert isinstance(X, np.ndarray)
        assert isinstance(Y, np.ndarray)
        assert len(X.shape) == 2
        assert len(Y.shape) == 2
        assert X.shape[0] == Y.shape[0]
        assert Y.shape[1] == 1

        time_start = time.time()

        X_ = X
        Y_ = Y
        time_all_ = []
        time_surrogate_ = []
        time_acq_ = []

        for ind_iter in range(0, self.num_iter):
            logger.info('Iteration %d', ind_iter + 1)
            time_iter_start = time.time()

            next_sample, dict_info = self.optimize_single_iteration(X_, Y_)

            next_samples = dict_info['next_points']
            acq_vals = dict_info['acquisitions']
            time_surrogate = dict_info['time_surrogate']
            time_acq = dict_info['time_acq']

            if self.debug:
                logger.debug('next_sample: %s', utils_logger.get_str_array(next_sample))

            next_sample = self._get_next_best_sample(next_sample, X_, next_samples, acq_vals)

            X_ = np.vstack((X_, next_sample))

            time_to_evaluate_start = time.time()
            Y_ = np.vstack((Y_, self.fun_target(next_sample)))
            time_to_evaluate_end = time.time()

            if self.debug:
                logger.debug('time consumed to evaluate: %.4f sec.',
                    time_to_evaluate_end - time_to_evaluate_start)

            time_iter_end = time.time()
            time_all_.append(time_iter_end - time_iter_start)
            time_surrogate_.append(time_surrogate)
            time_acq_.append(time_acq)

        time_end = time.time()

        if self.debug:
            logger.debug('overall time consumed in single BO round: %.4f sec.', time_end - time_start)

        time_all_ = np.array(time_all_)
        time_surrogate_ = np.array(time_surrogate_)
        time_acq_ = np.array(time_acq_)

        return X_, Y_, time_all_, time_surrogate_, time_acq_

    def optimize_with_initial_inputs(self,
        X: np.ndarray,
    ) -> constants.TYPING_TUPLE_FIVE_ARRAYS:
        """
        It returns the optimization results and times consumed, given
        inital inputs `X`.

        :param X: initial inputs. Shape: (n, d) or (n, m, d).
        :type X: numpy.ndarray

        :returns: a tuple of acquired samples, their function values, overall
            times consumed per iteration, time consumed in modeling Gaussian process
            regression, and time consumed in acquisition function optimization.
            Shape: ((n + `num_iter`, d), (n + `num_iter`, 1),
            (n + `num_iter`, ), (`num_iter`, ), (`num_iter`, )),
            or ((n + `num_iter`, m, d),
            (n + `num_iter`, m, 1), (n + `num_iter`, ), (`num_iter`, ),
            (`num_iter`, )).
        :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)

        :raises: AssertionError

        """

        assert isinstance(X, np.ndarray)
        assert len(X.shape) == 2

        Y = []
        time_initials = []
        for elem in X:
            time_initial_start = time.time()
            Y.append(self.fun_target(elem))
            time_initial_end = time.time()
            time_initials.append(time_initial_end - time_initial_start)
        time_initials = np.array(time_initials)

        Y = np.array(Y)
        Y = np.reshape(Y, (Y.shape[0], 1))

        X_, Y_, time_all_, time_surrogate_, time_acq_ \
            = self.optimize_with_all_initial_information(X, Y)

        time_all_ = np.concatenate((time_initials, time_all_))

        return X_, Y_, time_all_, time_surrogate_, time_acq_

    def print_info(self, num_init, seed):
        logger.info('====================')
        logger.info('range_X:\n%s', utils_logger.get_str_array(self.range_X))
        logger.info('num_init: %d', num_init)
        logger.info('num_iter: %d', self.num_iter)
        logger.info('str_surrogate: %s', self.str_surrogate)
        if self.str_surrogate in constants.ALLOWED_SURROGATE:
            logger.info('str_cov: %s', self.str_cov)
        logger.info('str_acq: %s', self.str_acq)
        logger.info('normalize_Y: %s', self.normalize_Y)
        if self.str_surrogate in constants.ALLOWED_SURROGATE:
            logger.info('use_ard: %s', self.use_ard)
        logger.info('str_initial_method_bo: %s', self.str_initial_method_bo)
        logger.info('str_sampling_method_ao: %s', self.str_sampling_method_ao)
        if self.str_surrogate in ['gp']:
            logger.info('str_optimizer_method_gp: %s', self.str_optimizer_method_gp)
        if self.str_surrogate in ['tp']:
            logger.info('str_optimizer_method_tp: %s', self.str_optimizer_method_tp)
        logger.info('str_optimizer_method_bo: %s', self.str_optimizer_method_bo)
        if self.str_surrogate in ['gp']:
            logger.info('str_mlm_method: %s', self.str_mlm_method)
            logger.info('str_modelselection_method: %s', self.str_modelselection_method)
        logger.info('num_samples_ao: %d', self.num_samples_ao)
        logger.info('seed: %s', seed)
        logger.info('debug: %s', self.debug)
        logger.info('====================')

    def optimize(self,
        num_init: int,
        seed: constants.TYPING_UNION_INT_NONE=None,
    ) -> constants.TYPING_TUPLE_FIVE_ARRAYS:
        """
        It returns the optimization results and times consumed, given
        the number of initial samples `num_init` and a random seed
        `seed`.

        :param num_init: the number of initial samples.
        :type num_init: int.
        :param seed: None, or a random seed.
        :type seed: NoneType or int., optional

        :returns: a tuple of acquired samples, their function values, overall
            times consumed per iteration, time consumed in modeling Gaussian process
            regression, and time consumed in acquisition function optimization.
            Shape: ((`num_init` + `num_iter`, d), (`num_init` + `num_iter`, 1),
            (`num_init` + `num_iter`, ), (`num_iter`, ), (`num_iter`, )),
            or ((`num_init` + `num_iter`, m, d),
            (`num_init` + `num_iter`, m, 1), (`num_init` + `num_iter`, ),
            (`num_iter`, ),
            (`num_iter`, )),
            where d is a dimensionality of the problem we are solving
            and m is a cardinality of sets.
        :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray)

        :raises: AssertionError

        """

        assert isinstance(num_init, int)
        assert isinstance(seed, (int, type(None)))
        assert num_init > 0

        self.print_info(num_init, seed)

        time_start = time.time()

        X_init = self.model_bo.get_initials(self.str_initial_method_bo, num_init, seed=seed)
        if self.debug:
            logger.debug('X_init:\n%s', utils_logger.get_str_array(X_init))

        X, Y, time_all, time_surrogate, time_acq = self.optimize_with_initial_inputs(X_init)

        time_end = time.time()

        if self.debug:
            logger.debug('overall time consumed including initializations: %.4f sec.',
                time_end - time_start)

        return X, Y, time_all, time_surrogate, time_acq



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 13, 2021
#
"""This file declares various default constants.
If you would like to see the details, check out
the Python script in the repository directly."""

import typing
import numpy as np


JITTER_ACQ = 1e-5
JITTER_COV = 1e-5
JITTER_LOG = 1e-7

TOLERANCE_DUPLICATED_ACQ = 1e-4

STR_SURROGATE = 'gp'
STR_SURROGATE_TREES = 'rf'
STR_OPTIMIZER_METHOD_GP = 'BFGS'
STR_OPTIMIZER_METHOD_TP = 'SLSQP'
STR_COV = 'matern52'
STR_BO_ACQ = 'ei'
STR_INITIALIZING_METHOD_BO = 'sobol'
STR_OPTIMIZER_METHOD_AO = 'L-BFGS-B'
STR_SAMPLING_METHOD_AO = 'sobol'
STR_OPTIMIZER_METHOD_AO_TREES = 'random_search'
STR_SAMPLING_METHOD_AO_TREES = 'uniform'
STR_MLM_METHOD = 'regular'
STR_MODELSELECTION_METHOD = 'ml'

NUM_GRIDS_AO = 50
NUM_SAMPLES_AO = 100
NUM_SAMPLES_AO_TREES = 5000

MULTIPLIER_ACQ = 10.0
MULTIPLIER_RESPONSE = 10.0

NORMALIZE_RESPONSE = True

USE_ARD = True
GP_NOISE = 1e-2
FIX_GP_NOISE = True
BOUND_UPPER_GP_NOISE = np.inf
RANGE_SIGNAL = [[1e-2, 1e3]]
RANGE_LENGTHSCALES = [[1e-2, 1e3]]
RANGE_NOISE = [[1e-3, 1e1]]
RANGE_DOF = [[2.00001, 200.0]]

TIME_PAUSE = 2.0
RANGE_SHADE = 1.96

NUM_DATA_PER_SPLIT_TREES = 100

ALLOWED_OPTIMIZER_METHOD_GP = [
    'BFGS',
    'L-BFGS-B',
    'Nelder-Mead',
    'SLSQP',
    'SLSQP-Bounded',
]
ALLOWED_OPTIMIZER_METHOD_TP = ['L-BFGS-B', 'SLSQP']
ALLOWED_OPTIMIZER_METHOD_BO = ['L-BFGS-B', 'DIRECT', 'CMA-ES']
ALLOWED_OPTIMIZER_METHOD_BO_TREES = ['random_search']

# INFO: Do not use _ (underscore) in base str_cov.
ALLOWED_COV_BASE = ['eq', 'se', 'matern32', 'matern52']
ALLOWED_COV_SET = ['set_' + str_cov for str_cov in ALLOWED_COV_BASE]
ALLOWED_COV = ALLOWED_COV_BASE + ALLOWED_COV_SET
ALLOWED_BO_ACQ = ['pi', 'ei', 'ucb', 'aei', 'pure_exploit', 'pure_explore']
ALLOWED_INITIALIZING_METHOD_BO = ['uniform', 'gaussian', 'sobol', 'halton']
ALLOWED_SAMPLING_METHOD = ALLOWED_INITIALIZING_METHOD_BO + ['grid']
ALLOWED_MLM_METHOD = ['regular', 'combined', 'converged']
ALLOWED_MODELSELECTION_METHOD = ['ml', 'loocv']
ALLOWED_SURROGATE = ['gp', 'tp']
ALLOWED_SURROGATE_TREES = ['rf']

COLORS = np.array([
    'red',
    'green',
    'blue',
    'orange',
    'olive',
    'purple',
    'darkred',
    'limegreen',
    'deepskyblue',
    'lightsalmon',
    'aquamarine',
    'navy',
    'rosybrown',
    'darkkhaki',
    'darkslategray',
])

MARKERS = np.array([
    '.',
    'x',
    '*',
    '+',
    '^',
    'v',
    '<',
    '>',
    'd',
    ',',
    '8',
    'h',
    '1',
    '2',
    '3',
])

TYPE_NONE = type(None)
TYPE_ARR = np.ndarray

TYPING_CALLABLE = typing.Callable
TYPING_LIST = typing.List
TYPING_TUPLE_DICT_BOOL = typing.Tuple[dict, bool]
TYPING_TUPLE_ARRAY_BOOL = typing.Tuple[TYPE_ARR, bool]
TYPING_TUPLE_ARRAY_DICT = typing.Tuple[TYPE_ARR, dict]
TYPING_TUPLE_ARRAY_FLOAT = typing.Tuple[TYPE_ARR, float]
TYPING_TUPLE_TWO_ARRAYS = typing.Tuple[TYPE_ARR, TYPE_ARR]
TYPING_TUPLE_TWO_ARRAYS_DICT = typing.Tuple[TYPE_ARR, TYPE_ARR, dict]
TYPING_TUPLE_TWO_FLOATS = typing.Tuple[float, float]
TYPING_TUPLE_THREE_ARRAYS = typing.Tuple[TYPE_ARR, TYPE_ARR, TYPE_ARR]
TYPING_TUPLE_FIVE_ARRAYS = typing.Tuple[TYPE_ARR, TYPE_ARR, TYPE_ARR, TYPE_ARR, TYPE_ARR]
TYPING_TUPLE_FLOAT_THREE_ARRAYS = typing.Tuple[float, TYPE_ARR, TYPE_ARR, TYPE_ARR]
TYPING_TUPLE_FLOAT_ARRAY = typing.Tuple[float, TYPE_ARR]

TYPING_UNION_INT_NONE = typing.Union[int, TYPE_NONE]
TYPING_UNION_INT_FLOAT = typing.Union[int, float]
TYPING_UNION_FLOAT_NONE = typing.Union[float, TYPE_NONE]
TYPING_UNION_FLOAT_TWO_FLOATS = typing.Union[float, TYPING_TUPLE_TWO_ARRAYS]
TYPING_UNION_FLOAT_FA = typing.Union[float, TYPING_TUPLE_FLOAT_ARRAY]
TYPING_UNION_ARRAY_NONE = typing.Union[TYPE_ARR, TYPE_NONE]
TYPING_UNION_ARRAY_FLOAT = typing.Union[TYPE_ARR, float]
TYPING_UNION_CALLABLE_NONE = typing.Union[TYPING_CALLABLE, TYPE_NONE]
TYPING_UNION_STR_NONE = typing.Union[str, TYPE_NONE]



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: September 24, 2020
#
"""These files are for implementing various wrappers."""

from bayeso.wrappers import wrappers_bo_function
from bayeso.wrappers import wrappers_bo_class


run_single_round_with_all_initial_information = wrappers_bo_function.run_single_round_with_all_initial_information
run_single_round_with_initial_inputs = wrappers_bo_function.run_single_round_with_initial_inputs
run_single_round = wrappers_bo_function.run_single_round

BayesianOptimization = wrappers_bo_class.BayesianOptimization



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: October 8, 2021
#
"""It defines a class of Bayesian optimization
with Student-:math:`t` process regression."""

import numpy as np
import time
from scipy.optimize import minimize
try:
    from scipydirect import minimize as directminimize
except: # pragma: no cover
    directminimize = None
try:
    import cma
except: # pragma: no cover
    cma = None

from bayeso.bo import base_bo
from bayeso import covariance
from bayeso import constants
from bayeso.tp import tp
from bayeso.tp import tp_kernel
from bayeso.utils import utils_bo
from bayeso.utils import utils_logger


class BOwTP(base_bo.BaseBO):
    """
    It is a Bayesian optimization class with Student-:math:`t` process regression.

    :param range_X: a search space. Shape: (d, 2).
    :type range_X: numpy.ndarray
    :param str_cov: the name of covariance function.
    :type str_cov: str., optional
    :param str_acq: the name of acquisition function.
    :type str_acq: str., optional
    :param normalize_Y: flag for normalizing outputs.
    :type normalize_Y: bool., optional
    :param use_ard: flag for automatic relevance determination.
    :type use_ard: bool., optional
    :param prior_mu: None, or prior mean function.
    :type prior_mu: NoneType, or callable, optional
    :param str_optimizer_method_tp: the name of optimization method for
        Student-:math:`t` process regression.
    :type str_optimizer_method_tp: str., optional
    :param str_optimizer_method_bo: the name of optimization method for
        Bayesian optimization.
    :type str_optimizer_method_bo: str., optional
    :param debug: flag for printing log messages.
    :type debug: bool., optional

    """

    def __init__(self, range_X: np.ndarray,
        str_cov: str=constants.STR_COV,
        str_acq: str=constants.STR_BO_ACQ,
        normalize_Y: bool=constants.NORMALIZE_RESPONSE,
        use_ard: bool=constants.USE_ARD,
        prior_mu: constants.TYPING_UNION_CALLABLE_NONE=None,
        str_optimizer_method_tp: str=constants.STR_OPTIMIZER_METHOD_TP,
        str_optimizer_method_bo: str=constants.STR_OPTIMIZER_METHOD_AO,
        debug: bool=False
    ):
        """
        Constructor method

        """

        assert isinstance(range_X, np.ndarray)
        assert isinstance(str_cov, str)
        assert isinstance(str_acq, str)
        assert isinstance(normalize_Y, bool)
        assert isinstance(use_ard, bool)
        assert isinstance(str_optimizer_method_bo, str)
        assert isinstance(str_optimizer_method_tp, str)
        assert isinstance(debug, bool)
        assert callable(prior_mu) or prior_mu is None
        assert len(range_X.shape) == 2
        assert range_X.shape[1] == 2
        assert (range_X[:, 0] <= range_X[:, 1]).all()
        assert str_cov in constants.ALLOWED_COV
        assert str_acq in constants.ALLOWED_BO_ACQ
        assert str_optimizer_method_tp in constants.ALLOWED_OPTIMIZER_METHOD_TP
        assert str_optimizer_method_bo in constants.ALLOWED_OPTIMIZER_METHOD_BO

        str_surrogate = 'tp'
        assert str_surrogate in constants.ALLOWED_SURROGATE

        super().__init__(range_X, str_surrogate, str_acq, str_optimizer_method_bo, normalize_Y, debug)

        self.str_cov = str_cov
        self.use_ard = use_ard
        self.str_optimizer_method_tp = str_optimizer_method_tp
        self.prior_mu = prior_mu

        self.historical_hyps = []

    def _optimize(self, fun_negative_acquisition: constants.TYPING_CALLABLE,
        str_sampling_method: str,
        num_samples: int
    ) -> constants.TYPING_TUPLE_TWO_ARRAYS:
        """
        It optimizes `fun_negative_function` with `self.str_optimizer_method_bo`.
        `num_samples` examples are determined by `str_sampling_method`, to
        start acquisition function optimization.

        :param fun_objective: negative acquisition function.
        :type fun_objective: callable
        :param str_sampling_method: the name of sampling method.
        :type str_sampling_method: str.
        :param num_samples: the number of samples.
        :type num_samples: int.

        :returns: tuple of next point to evaluate and all candidates
            determined by acquisition function optimization.
            Shape: ((d, ), (`num_samples`, d)).
        :rtype: (numpy.ndarray, numpy.ndarray)

        """

        list_next_point = []
        if self.str_optimizer_method_bo == 'L-BFGS-B':
            list_bounds = self._get_bounds()
            initials = self.get_samples(str_sampling_method,
                fun_objective=fun_negative_acquisition,
                num_samples=num_samples)

            for arr_initial in initials:
                next_point = minimize(
                    fun_negative_acquisition,
                    x0=arr_initial,
                    bounds=list_bounds,
                    method=self.str_optimizer_method_bo,
                    options={'disp': False}
                )
                next_point_x = next_point.x
                list_next_point.append(next_point_x)
                if self.debug:
                    self.logger.debug('acquired sample: %s', utils_logger.get_str_array(next_point_x))
        elif self.str_optimizer_method_bo == 'DIRECT': # pragma: no cover
            self.logger.debug('num_samples is ignored.')

            list_bounds = self._get_bounds()
            next_point = directminimize(
                fun_negative_acquisition,
                bounds=list_bounds,
                maxf=88888,
            )
            next_point_x = next_point.x
            list_next_point.append(next_point_x)
        elif self.str_optimizer_method_bo == 'CMA-ES':
            self.logger.debug('num_samples is ignored.')

            list_bounds = self._get_bounds()
            list_bounds = np.array(list_bounds)
            def fun_wrapper(f):
                def g(bx):
                    return f(bx)[0]
                return g
            initials = self.get_samples(str_sampling_method,
                fun_objective=fun_negative_acquisition, num_samples=1)
            cur_sigma0 = np.mean(list_bounds[:, 1] - list_bounds[:, 0]) / 4.0
            next_point_x = cma.fmin(fun_wrapper(fun_negative_acquisition),
                initials[0], cur_sigma0,
                options={
                    'bounds': [list_bounds[:, 0], list_bounds[:, 1]],
                    'verbose': -1, 'maxfevals': 1e5
                })[0]
            list_next_point.append(next_point_x)

        next_points = np.array(list_next_point)
        next_point = utils_bo.get_best_acquisition_by_evaluation(
            next_points, fun_negative_acquisition)[0]
        return next_point, next_points

    def compute_posteriors(self,
        X_train: np.ndarray, Y_train: np.ndarray,
        X_test: np.ndarray, cov_X_X: np.ndarray,
        inv_cov_X_X: np.ndarray, hyps: dict
    ) -> np.ndarray:
        """
        It returns posterior mean and standard deviation over `X_test`.

        :param X_train: inputs. Shape: (n, d) or (n, m, d).
        :type X_train: numpy.ndarray
        :param Y_train: outputs. Shape: (n, 1).
        :type Y_train: numpy.ndarray
        :param X_test: inputs. Shape: (l, d) or (l, m, d).
        :type X_test: numpy.ndarray
        :param cov_X_X: kernel matrix over `X_train`. Shape: (n, n).
        :type cov_X_X: numpy.ndarray
        :param inv_cov_X_X: kernel matrix inverse over `X_train`. Shape: (n, n).
        :type inv_cov_X_X: numpy.ndarray
        :param hyps: dictionary of hyperparameters for Gaussian process.
        :type hyps: dict.

        :returns: posterior mean and standard deviation over `X_test`. Shape: ((l, ), (l, )).
        :rtype: (numpy.ndarray, numpy.ndarray)

        :raises: AssertionError

        """

        assert isinstance(X_train, np.ndarray)
        assert isinstance(Y_train, np.ndarray)
        assert isinstance(X_test, np.ndarray)
        assert isinstance(cov_X_X, np.ndarray)
        assert isinstance(inv_cov_X_X, np.ndarray)
        assert isinstance(hyps, dict)
        assert len(X_train.shape) == 2 or len(X_train.shape) == 3
        assert len(Y_train.shape) == 2
        assert len(X_test.shape) == 2 or len(X_test.shape) == 3
        assert len(cov_X_X.shape) == 2
        assert len(inv_cov_X_X.shape) == 2
        assert Y_train.shape[1] == 1
        assert X_train.shape[0] == Y_train.shape[0]
        if len(X_train.shape) == 2:
            assert X_test.shape[1] == X_train.shape[1] == self.num_dim
        else:
            assert X_test.shape[2] == X_train.shape[2] == self.num_dim
        assert cov_X_X.shape[0] == cov_X_X.shape[1] == X_train.shape[0]
        assert inv_cov_X_X.shape[0] == inv_cov_X_X.shape[1] == X_train.shape[0]

        _, pred_mean, pred_std, _ = tp.predict_with_cov(
            X_train, Y_train, X_test,
            cov_X_X, inv_cov_X_X, hyps, str_cov=self.str_cov,
            prior_mu=self.prior_mu, debug=self.debug
        )

        pred_mean = np.squeeze(pred_mean, axis=1)
        pred_std = np.squeeze(pred_std, axis=1)

        return pred_mean, pred_std

    def compute_acquisitions(self, X: np.ndarray,
        X_train: np.ndarray, Y_train: np.ndarray,
        cov_X_X: np.ndarray, inv_cov_X_X: np.ndarray, hyps: dict
    ) -> np.ndarray:
        """
        It computes acquisition function values over 'X',
        where `X_train`, `Y_train`, `cov_X_X`, `inv_cov_X_X`, and `hyps`
        are given.

        :param X: inputs. Shape: (l, d) or (l, m, d).
        :type X: numpy.ndarray
        :param X_train: inputs. Shape: (n, d) or (n, m, d).
        :type X_train: numpy.ndarray
        :param Y_train: outputs. Shape: (n, 1).
        :type Y_train: numpy.ndarray
        :param cov_X_X: kernel matrix over `X_train`. Shape: (n, n).
        :type cov_X_X: numpy.ndarray
        :param inv_cov_X_X: kernel matrix inverse over `X_train`. Shape: (n, n).
        :type inv_cov_X_X: numpy.ndarray
        :param hyps: dictionary of hyperparameters.
        :type hyps: dict.

        :returns: acquisition function values over `X`. Shape: (l, ).
        :rtype: numpy.ndarray

        :raises: AssertionError

        """

        assert isinstance(X, np.ndarray)
        assert isinstance(X_train, np.ndarray)
        assert isinstance(Y_train, np.ndarray)
        assert isinstance(cov_X_X, np.ndarray)
        assert isinstance(inv_cov_X_X, np.ndarray)
        assert isinstance(hyps, dict)
        assert len(X.shape) == 1 or len(X.shape) == 2 or len(X.shape) == 3
        assert len(X_train.shape) == 2 or len(X_train.shape) == 3
        assert len(Y_train.shape) == 2
        assert len(cov_X_X.shape) == 2
        assert len(inv_cov_X_X.shape) == 2
        assert Y_train.shape[1] == 1
        assert X_train.shape[0] == Y_train.shape[0]

        if len(X.shape) == 1:
            X = np.atleast_2d(X)

        if len(X_train.shape) == 2:
            assert X.shape[1] == X_train.shape[1] == self.num_dim
        else:
            assert X.shape[2] == X_train.shape[2] == self.num_dim

        assert cov_X_X.shape[0] == cov_X_X.shape[1] == X_train.shape[0]
        assert inv_cov_X_X.shape[0] == inv_cov_X_X.shape[1] == X_train.shape[0]

        fun_acquisition = utils_bo.choose_fun_acquisition(self.str_acq, hyps.get('noise', None))

        pred_mean, pred_std = self.compute_posteriors(
            X_train, Y_train, X,
            cov_X_X, inv_cov_X_X, hyps
        )

        acquisitions = fun_acquisition(
            pred_mean=pred_mean, pred_std=pred_std, Y_train=Y_train
        )
        acquisitions *= constants.MULTIPLIER_ACQ

        return acquisitions

    def optimize(self, X_train: np.ndarray, Y_train: np.ndarray,
        str_sampling_method: str=constants.STR_SAMPLING_METHOD_AO,
        num_samples: int=constants.NUM_SAMPLES_AO,
    ) -> constants.TYPING_TUPLE_ARRAY_DICT:
        """
        It computes acquired example, candidates of acquired examples,
        acquisition function values for the candidates, covariance matrix,
        inverse matrix of the covariance matrix, hyperparameters optimized,
        and execution times.

        :param X_train: inputs. Shape: (n, d) or (n, m, d).
        :type X_train: numpy.ndarray
        :param Y_train: outputs. Shape: (n, 1).
        :type Y_train: numpy.ndarray
        :param str_sampling_method: the name of sampling method for
            acquisition function optimization.
        :type str_sampling_method: str., optional
        :param num_samples: the number of samples.
        :type num_samples: int., optional

        :returns: acquired example and dictionary of information. Shape: ((d, ), dict.).
        :rtype: (numpy.ndarray, dict.)

        :raises: AssertionError

        """

        assert isinstance(X_train, np.ndarray)
        assert isinstance(Y_train, np.ndarray)
        assert isinstance(str_sampling_method, str)
        assert isinstance(num_samples, int)
        assert len(X_train.shape) == 2
        assert len(Y_train.shape) == 2
        assert Y_train.shape[1] == 1
        assert X_train.shape[0] == Y_train.shape[0]
        assert X_train.shape[1] == self.num_dim
        assert num_samples > 0
        assert str_sampling_method in constants.ALLOWED_SAMPLING_METHOD

        time_start = time.time()

        if self.normalize_Y and np.max(Y_train) != np.min(Y_train):
            Y_train = (Y_train - np.min(Y_train)) / (np.max(Y_train) - np.min(Y_train)) \
                * constants.MULTIPLIER_RESPONSE

        time_start_surrogate = time.time()

        cov_X_X, inv_cov_X_X, hyps = tp_kernel.get_optimized_kernel(
            X_train, Y_train,
            self.prior_mu, self.str_cov,
            str_optimizer_method=self.str_optimizer_method_tp,
            use_ard=self.use_ard,
            debug=self.debug
        )

        self.historical_hyps.append(hyps)

        time_end_surrogate = time.time()

        time_start_acq = time.time()
        fun_negative_acquisition = lambda X_test: -1.0 * self.compute_acquisitions(
            X_test, X_train, Y_train, cov_X_X, inv_cov_X_X, hyps
        )
        next_point, next_points = self._optimize(fun_negative_acquisition,
            str_sampling_method=str_sampling_method, num_samples=num_samples)
        time_end_acq = time.time()

        acquisitions = fun_negative_acquisition(next_points)
        time_end = time.time()

        dict_info = {
            'next_points': next_points,
            'acquisitions': acquisitions,
            'cov_X_X': cov_X_X,
            'inv_cov_X_X': inv_cov_X_X,
            'hyps': hyps,
            'time_surrogate': time_end_surrogate - time_start_surrogate,
            'time_acq': time_end_acq - time_start_acq,
            'time_overall': time_end - time_start,
        }

        if self.debug:
            self.logger.debug('overall time consumed to acquire: %.4f sec.', time_end - time_start)

        return next_point, dict_info



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: September 24, 2020
#
"""It is utilities for loggers."""

import logging
import numpy as np

from bayeso.utils import utils_common


@utils_common.validate_types
def get_logger(str_name: str) -> logging.Logger:
    """
    It returns a logger to record the messages generated in our package.

    :param str_name: a logger name.
    :type str_name: str.

    :returns: a logger.
    :rtype: logging.Logger

    :raises: AssertionError

    """

    assert isinstance(str_name, str)

    logger = logging.getLogger(str_name)
    logger.setLevel(logging.DEBUG)

    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    formatter = logging.Formatter('[%(levelname)s-%(name)s-%(asctime)s] %(message)s',
        datefmt='%m/%d/%Y-%H:%M:%S')
    ch.setFormatter(formatter)

    logger.addHandler(ch)
    return logger

@utils_common.validate_types
def get_str_array_1d(arr: np.ndarray) -> str:
    """
    It converts a one-dimensional array into string.

    :param arr: an array to be converted.
    :type arr: numpy.ndarray

    :returns: a string.
    :rtype: str.

    :raises: AssertionError

    """

    assert isinstance(arr, np.ndarray)
    assert len(arr.shape) == 1

    list_str = []

    for elem in arr:
        if isinstance(elem, float):
            elem_ = '{:.3f}'.format(elem)
        else:
            elem_ = '{}'.format(elem)

        list_str.append(elem_)

    str_arr = ', '.join(list_str)
    str_arr = '[' + str_arr + ']'
    return str_arr

@utils_common.validate_types
def get_str_array_2d(arr: np.ndarray) -> str:
    """
    It converts a two-dimensional array into string.

    :param arr: an array to be converted.
    :type arr: numpy.ndarray

    :returns: a string.
    :rtype: str.

    :raises: AssertionError

    """

    assert isinstance(arr, np.ndarray)
    assert len(arr.shape) == 2

    list_str = [get_str_array_1d(elem) for elem in arr]

    str_arr = ',\n'.join(list_str)
    str_arr = '[' + str_arr + ']'
    return str_arr

@utils_common.validate_types
def get_str_array_3d(arr: np.ndarray) -> str:
    """
    It converts a three-dimensional array into string.

    :param arr: an array to be converted.
    :type arr: numpy.ndarray

    :returns: a string.
    :rtype: str.

    :raises: AssertionError

    """

    assert isinstance(arr, np.ndarray)
    assert len(arr.shape) == 3

    list_str = [get_str_array_2d(elem) for elem in arr]

    str_arr = ',\n'.join(list_str)
    str_arr = '[' + str_arr + ']'
    return str_arr

@utils_common.validate_types
def get_str_array(arr: np.ndarray) -> str:
    """
    It converts an array into string. It can take one-dimensional,
    two-dimensional, and three-dimensional arrays.

    :param arr: an array to be converted.
    :type arr: numpy.ndarray

    :returns: a string.
    :rtype: str.

    :raises: AssertionError

    """

    assert isinstance(arr, np.ndarray)
    len_arr = len(arr.shape)

    if len_arr == 1:
        str_arr = get_str_array_1d(arr)
    elif len_arr == 2:
        str_arr = get_str_array_2d(arr)
    elif len_arr == 3:
        str_arr = get_str_array_3d(arr)
    else:
        raise NotImplementedError('invalid len_arr.')

    return str_arr

@utils_common.validate_types
def get_str_hyps(hyps: dict) -> str:
    """
    It converts a dictionary of hyperparameters into string.

    :param hyps: a hyperparameter dictionary to be converted.
    :type hyps: dict.

    :returns: a string.
    :rtype: str.

    :raises: AssertionError

    """

    assert isinstance(hyps, dict)

    list_str = []

    for key, val in hyps.items():
        if isinstance(val, np.ndarray):
            str_val = get_str_array(val)
        elif isinstance(val, float):
            str_val = '{:.3f}'.format(val)
        else:
            str_val = '{}'.format(val)

        list_str.append("'{}'".format(key) + ': ' + str_val)

    str_hyps = ', '.join(list_str)
    str_hyps = '{' + str_hyps + '}'
    return str_hyps



#
# author: Jungtaek Kim (jtkim@postech.ac.kr)
# last updated: September 24, 2020
#
"""It is utilities for common features."""

import functools
import numpy as np

from bayeso import constants


def validate_types(func: constants.TYPING_CALLABLE) -> constants.TYPING_CALLABLE:
    """
    It is a decorator for validating the number of types, which are declared for typing.

    :param func: an original function.
    :type func: callable

    :returns: a callable decorator.
    :rtype: callable

    :raises: AssertionError

    """

    annos = func.__annotations__
    assert len(annos) == func.__code__.co_argcount + 1
#    arg_names = func.__code__.co_varnames[:func.__code__.co_argcount]

    @functools.wraps(func)
    def _validate_types(*args, **kwargs):
        return func(*args, **kwargs)

    return _validate_types

@validate_types
def get_grids(ranges: np.ndarray, num_grids: int) -> np.ndarray:
    """
    It returns grids of given `ranges`, where each of dimension has `num_grids` partitions.

    :param ranges: ranges. Shape: (d, 2).
    :type ranges: numpy.ndarray
    :param num_grids: the number of partitions per dimension.
    :type num_grids: int.

    :returns: grids of given `ranges`. Shape: (`num_grids`:math:`^{\\text{d}}`, d).
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(ranges, np.ndarray)
    assert isinstance(num_grids, int)
    assert len(ranges.shape) == 2
    assert ranges.shape[1] == 2
    assert (ranges[:, 0] <= ranges[:, 1]).all()

    list_grids = []
    for range_ in ranges:
        list_grids.append(np.linspace(range_[0], range_[1], num_grids))
    list_grids_mesh = list(np.meshgrid(*list_grids))
    list_grids = []
    for elem in list_grids_mesh:
        list_grids.append(elem.flatten(order='C'))
    arr_grids = np.vstack(tuple(list_grids))
    arr_grids = arr_grids.T
    return arr_grids

@validate_types
def get_minimum(Y_all: np.ndarray, num_init: int) -> constants.TYPING_TUPLE_THREE_ARRAYS:
    """
    It returns accumulated minima at each iteration, their arithmetic means
    over rounds, and their standard deviations over rounds, which is widely
    used in Bayesian optimization community.

    :param Y_all: historical function values. Shape: (r, t) where r is the
        number of Bayesian optimization rounds and t is the number of
        iterations including initial points for each round. For example,
        if we run 50 iterations with 5 initial examples and repeat this
        procedure 3 times, r would be 3 and t would be 55 (= 50 + 5).
    :type Y_all: numpy.ndarray
    :param num_init: the number of initial points.
    :type num_init: int.

    :returns: tuple of accumulated minima, their arithmetic means over
        rounds, and their standard deviations over rounds.
        Shape: ((r, t - `num_init` + 1), (t - `num_init` + 1, ), (t - `num_init` + 1, )).
    :rtype: (numpy.ndarray, numpy.ndarray, numpy.ndarray)

    :raises: AssertionError

    """

    assert isinstance(Y_all, np.ndarray)
    assert isinstance(num_init, int)
    assert len(Y_all.shape) == 2
    assert Y_all.shape[1] > num_init

    list_minima = []

    for by in Y_all:
        minimum_best = np.inf
        list_minima_ = []
        for y in by[:num_init]:
            if minimum_best > y:
                minimum_best = y
        list_minima_.append(minimum_best)
        for y in by[num_init:]:
            if minimum_best > y:
                minimum_best = y
            list_minima_.append(minimum_best)
        list_minima.append(list_minima_)

    minima = np.array(list_minima)
    mean_minima = np.mean(minima, axis=0)
    std_minima = np.std(minima, axis=0)

    return minima, mean_minima, std_minima

@validate_types
def get_time(time_all: np.ndarray, num_init: int, include_init: bool) -> np.ndarray:
    """
    It returns the means of accumulated execution times over rounds.

    :param time_all: execution times for all Bayesian optimization rounds.
        Shape: (r, t) where r is the number of Bayesian optimization rounds
        and t is the number of iterations (including initial points if
        `include_init` is True, or excluding them if `include_init` is
        False) for each round.

    :type time_all: numpy.ndarray
    :param num_init: the number of initial points. If `include_init` is
        False, it is ignored even if it is provided.
    :type num_init: int.
    :param include_init: flag for describing whether execution times to
        observe initial examples have been included or not.
    :type include_init: bool.

    :returns: arithmetic means of accumulated execution times over rounds.
        Shape: (t - `num_init`, ) if `include_init` is True. (t, ), otherwise.
    :rtype: numpy.ndarray

    :raises: AssertionError

    """

    assert isinstance(time_all, np.ndarray)
    assert isinstance(num_init, int)
    assert isinstance(include_init, bool)
    assert len(time_all.shape) == 2
    if include_init:
        assert time_all.shape[1] > num_init

    list_time = []
    for time_ in time_all:
        list_time_ = np.array([0.0])

        if include_init:
            time_ = time_[num_init:]
        list_time_ = np.concatenate((list_time_, np.cumsum(time_)))
        list_time.append(list_time_)
    list_time = np.array(list_time)

    return np.mean(list_time, axis=0)
